#!/usr/bin/python3
"""Archive messages."""
from functools import lru_cache
from typing import List, Tuple, Type, cast

from sqlalchemy.orm import Session, class_mapper
from sqlalchemy_utils import get_type

from messenger.commands.utils import Verbosity, echo, messenger_command, type_option
from messenger.constants import ArchiveMessageStatus, TransportMethod
from messenger.database import atomic
from messenger.models import ArchiveRecord, Record


@lru_cache()
def _get_archived_columns(method: TransportMethod) -> Tuple[str, ...]:
    """Return columns copied to archive model."""
    shared_columns = set(class_mapper(method.model_cls).columns.keys()).intersection(
        class_mapper(method.archive_cls).columns.keys())
    return tuple(c for c in shared_columns if c != 'references')


def _archive_model(method: TransportMethod, session: Session, model: Record) -> None:
    """Archive single message."""
    # mypy gets confused by lru_cache, see https://github.com/python/mypy/issues/5858
    archive_cls = cast(Type[ArchiveRecord], method.archive_cls)
    data = {a: getattr(model, a) for a in _get_archived_columns(method)}
    archived = archive_cls(**data)
    reference_cls = get_type(archive_cls.references)
    for reference in model.references:
        arch_ref = reference_cls(id=reference.id, create_datetime=archived.create_datetime,
                                 message_id=reference.message_id, type=reference.type, value=reference.value)
        archived.references.append(arch_ref)
    session.add(archived)
    for reference in model.references:
        session.delete(reference)
    session.delete(model)


def _archive(method: TransportMethod) -> None:
    """Archive messages of single method."""
    archived = 0
    while True:
        with atomic() as session:
            # mypy gets confused by lru_cache, see https://github.com/python/mypy/issues/5858
            model_cls = cast(Record, method.model_cls)
            # Filter by status & archive
            query = session.query(model_cls).filter_by(archive=True).filter(
                model_cls.status.in_(ArchiveMessageStatus))
            # Fetch one message at a time.
            # Skip any message locked in other transaction and lock the selected message for updates.
            model = query.with_for_update(skip_locked=True).first()
            if model is None:
                # No more messages. Quit.
                break
            _archive_model(method, session, model)
            archived += 1
            echo('.', Verbosity.DETAIL, nl=False)
    # Print newline between progress and summary.
    echo('', Verbosity.DETAIL)
    echo('Archived {} {}s.'.format(archived, method.value.lower()), Verbosity.INFO)


@messenger_command()
@type_option
def main(types: List[TransportMethod]) -> None:
    """Archive messages."""
    for method in types:
        _archive(method)


if __name__ == '__main__':
    main()
