#!/usr/bin/python3
"""Print messenger statistics over defined period."""
import sys
from datetime import datetime, timedelta
from typing import List, Type, cast

import click
from pytz import UTC

from messenger.commands.utils import messenger_command, type_option
from messenger.constants import MessageStatus, TransportMethod
from messenger.database import atomic
from messenger.models import Record

OUTPUT = ('{method.value}s per last {period} seconds:\n'
          'Created: {created}\n'
          'Sent: {sent}\n'
          'Failed: {failed}\n'
          'Errors: {errors}\n'
          'Pending total: {pending}\n')


@messenger_command()
@type_option
@click.option('--period', type=int, default=3600, show_default=True,
              help="Set length of a monitored period in seconds.")
def main(types: List[TransportMethod], period: int) -> None:
    """Print messenger statistics over defined period."""
    cutoff = datetime.now(UTC) - timedelta(seconds=period)
    for method in types:
        with atomic() as session:
            # mypy gets confused by lru_cache, see https://github.com/python/mypy/issues/5858
            model_cls = cast(Type[Record], method.model_cls)
            query = session.query(model_cls)
            stats = {
                'method': method,
                'period': period,
                'created': query.filter(model_cls.create_datetime >= cutoff).count(),
                'pending': query.filter_by(status=MessageStatus.PENDING).count(),
                'sent': query.filter(model_cls.send_datetime >= cutoff).filter_by(status=MessageStatus.SENT).count(),
                'failed': query.filter(model_cls.update_datetime >= cutoff).filter_by(
                    status=MessageStatus.FAILED).count(),
                'errors': query.filter(model_cls.update_datetime >= cutoff).filter_by(
                    status=MessageStatus.ERROR).count(),
            }
        sys.stdout.write(OUTPUT.format(**stats))


if __name__ == '__main__':
    main()
