"""Command utilities."""
from enum import IntEnum, unique
from functools import wraps
from typing import Any, Callable, List, Union

import click
from click import echo as _echo
from typing_extensions import Literal

import messenger
from messenger.constants import TransportMethod
from messenger.settings import SETTINGS


@unique
class Verbosity(IntEnum):
    """Verbosity levels."""

    SILENT = 0
    INFO = 1
    DETAIL = 2
    DEBUG = 3


def echo(message: str, verbosity: Verbosity, *, nl: bool = True, err: bool = False) -> None:
    """Print message with respect to verbosity level."""
    if verbosity <= SETTINGS['_verbosity']:
        _echo(message, nl=nl, err=err)


def messenger_command(command: Callable = click.command) -> Callable:
    """Turn function into a messenger command.

    Arguments:
        command: A click `command` callable to attach command to a group.
    """
    def decorator(func: Callable) -> Callable:
        # This decorator turns the function into a click command. To keep a reasonable order of options, that is
        # end usage with --connection, --config, --verbosity, --version and --help in this order,
        # we need to keep this order of click decorators.
        @click.version_option(version=messenger.__version__, message="fred-messenger %(version)s")
        @click.option('-v', '--verbosity', type=int, default=1, show_default=True, help="Set verbosity level.")
        @click.option('--config', type=click.Path(exists=True, dir_okay=False), help="Set custom config file.")
        @click.option('--connection', help="Set custom database connection string.")
        @command()
        @wraps(func)
        def wrapped(*, config: str, connection: str, verbosity: int, **kwargs) -> Any:
            SETTINGS.load(config)
            SETTINGS['_verbosity'] = verbosity
            if connection:
                SETTINGS['db_connection'] = connection
            return func(**kwargs)

        return wrapped
    return decorator


def _parse_type(ctx: click.Context, param: click.Parameter, value: Union[Literal['all'], TransportMethod]
                ) -> List[TransportMethod]:
    """Return selected type as a collection of transport methods."""
    if value == 'all':
        return list(TransportMethod)
    else:
        # Only TransportMethod values and 'all' are allowed choices.
        assert isinstance(value, TransportMethod)  # nosec
        return [value]


def type_option(func: Callable) -> Callable:
    """Add --type option to a command."""
    # Use `types` as the name of the variable.
    option = click.option('--type', 'types', type=click.Choice(['all'] + list(TransportMethod)), default='all',
                          show_default=True, callback=_parse_type, help="Run only service with selected type.")
    return option(func)
