from datetime import datetime
from unittest import TestCase
from uuid import UUID

from click.testing import CliRunner, Result
from freezegun import freeze_time
from pytz import UTC

from messenger.commands import monitor
from messenger.constants import MessageStatus, TransportMethod
from messenger.database import get_session_cls
from messenger.models import Address
from messenger.settings import SETTINGS

from .utils import make_email, make_letter, make_sms, override_settings


@freeze_time('2020-02-01 14:00:00')
@override_settings(db_connection=SETTINGS['test_db_connection'])
class MonitorMainTest(TestCase):
    recipient = ['kryten@example.com']
    subject_template = 'subject.txt'
    body_template = 'body.txt'
    recent_time = datetime(2020, 2, 1, 13, 30, tzinfo=UTC)
    non_recent_time = datetime(2020, 2, 1, 12, 30, tzinfo=UTC)

    def setUp(self):
        self.session = get_session_cls()()
        self.runner = CliRunner()

    def tearDown(self):
        self.session.close()

    def assertOutputContains(self, result: Result, text: str) -> None:
        """Check if result output contains the expected text."""
        self.assertIn(text, result.stdout)

    def _test_monitor(self, created: int = 0, pending: int = 0, sent: int = 0, failed: int = 0, errors: int = 0):
        with override_settings():
            result = self.runner.invoke(monitor.main, ['--type', 'email'])

        expected = monitor.OUTPUT.format(method=TransportMethod.EMAIL, period=3600, created=created, pending=pending,
                                         sent=sent, failed=failed, errors=errors)
        self.assertOutputContains(result, expected)

    def test_main_no_emails(self):
        self._test_monitor()

    def test_main_new(self):
        make_email(self.session, recipient=self.recipient, create_datetime=self.recent_time,
                   subject_template=self.subject_template, body_template=self.body_template)

        self._test_monitor(created=1, pending=1)

    def test_main_sent(self):
        make_email(self.session, recipient=self.recipient, create_datetime=self.non_recent_time,
                   status=MessageStatus.SENT, send_datetime=self.recent_time,
                   subject_template=self.subject_template, body_template=self.body_template)

        self._test_monitor(sent=1)

    def test_main_failed(self):
        make_email(self.session, recipient=self.recipient, create_datetime=self.non_recent_time,
                   update_datetime=self.recent_time, status=MessageStatus.FAILED,
                   subject_template=self.subject_template, body_template=self.body_template)

        self._test_monitor(failed=1)

    def test_main_errors(self):
        make_email(self.session, recipient=self.recipient, create_datetime=self.non_recent_time,
                   update_datetime=self.recent_time, status=MessageStatus.ERROR,
                   subject_template=self.subject_template, body_template=self.body_template)

        self._test_monitor(errors=1)

    def test_main_pending(self):
        # Test pending emails are counted regardless of their age.
        make_email(self.session, recipient=self.recipient, create_datetime=self.non_recent_time,
                   subject_template=self.subject_template, body_template=self.body_template)

        self._test_monitor(pending=1)

    def test_main_custom_period(self):
        make_email(self.session, recipient=self.recipient, create_datetime=self.recent_time,
                   subject_template=self.subject_template, body_template=self.body_template)
        make_email(self.session, recipient=self.recipient, create_datetime=self.non_recent_time,
                   subject_template=self.subject_template, body_template=self.body_template)

        with override_settings():
            result = self.runner.invoke(monitor.main, ['--period', '7200', '--type', 'email'])

        expected = monitor.OUTPUT.format(method=TransportMethod.EMAIL, period=7200, created=2, pending=2, sent=0,
                                         failed=0, errors=0)
        self.assertOutputContains(result, expected)

    def test_types(self):
        make_email(self.session, recipient=self.recipient, create_datetime=self.recent_time,
                   subject_template=self.subject_template, body_template=self.body_template)
        make_sms(self.session, recipient='+1234', send_datetime=self.recent_time, status=MessageStatus.SENT,
                 body_template=self.body_template)
        make_letter(self.session, recipient=Address(name='Kryten'), update_datetime=self.recent_time,
                    status=MessageStatus.ERROR, file=UUID(int=42))
        with override_settings():
            result = self.runner.invoke(monitor.main)

        expected = (
            monitor.OUTPUT.format(method=TransportMethod.EMAIL, period=3600, created=1, pending=1, sent=0, failed=0,
                                  errors=0),
            monitor.OUTPUT.format(method=TransportMethod.SMS, period=3600, created=1, pending=0, sent=1, failed=0,
                                  errors=0),
            monitor.OUTPUT.format(method=TransportMethod.LETTER, period=3600, created=1, pending=0, sent=0, failed=0,
                                  errors=1),
        )
        self.assertOutputContains(result, ''.join(expected))
