from concurrent.futures import ThreadPoolExecutor
from typing import Dict, List
from unittest import TestCase
from unittest.mock import call, patch

from click.testing import CliRunner
from grpc._utilities import DictionaryGenericHandler
from testfixtures import Comparison

from messenger.commands import server
from messenger.settings import SETTINGS

from .utils import override_settings


@override_settings(db_connection=SETTINGS['test_db_connection'])
class ServiceMainTest(TestCase):
    def setUp(self):
        self.runner = CliRunner()

    def _test_main(self, argv: List[str], services: List[str] = None, port: int = 50051, options: Dict = None):
        services = services or ['Fred.Messenger.Api.Email.EmailMessenger', 'Fred.Messenger.Api.Sms.SmsMessenger',
                                'Fred.Messenger.Api.Letter.LetterMessenger']
        with patch('messenger.commands.server.sleep', side_effect=KeyboardInterrupt):
            with patch('messenger.commands.server.grpc') as grpc_mock:
                self.runner.invoke(server.main, argv)

        calls = [call.server(Comparison(ThreadPoolExecutor), options=options)]
        calls += len(services) * [call.server().add_generic_rpc_handlers((Comparison(DictionaryGenericHandler), ))]
        calls += [call.server().add_insecure_port('[::]:{}'.format(port)),
                  call.server().start(),
                  call.server().stop(0)]
        self.assertEqual(grpc_mock.mock_calls, calls)
        self.assertEqual([c[1][0][0]._name for c in grpc_mock.mock_calls[1:1 + len(services)]], services)

    def test_main(self):
        self._test_main([])

    def test_main_email(self):
        services = ['Fred.Messenger.Api.Email.EmailMessenger']
        self._test_main(['--type', 'email'], services)

    def test_main_sms(self):
        services = ['Fred.Messenger.Api.Sms.SmsMessenger']
        self._test_main(['--type', 'sms'], services)

    def test_main_letter(self):
        services = ['Fred.Messenger.Api.Letter.LetterMessenger']
        self._test_main(['--type', 'letter'], services)

    def test_port(self):
        self._test_main(['--port', '12345'], port=12345)

    def test_grpc_options(self):
        self._test_main(['--grpc-options', 'name=Kryten'], options={'name': 'Kryten'})
