import tempfile
from typing import List
from unittest import TestCase

import click
import yaml
from click.testing import CliRunner
from testfixtures import OutputCapture

import messenger
from messenger.commands.utils import Verbosity, echo, messenger_command, type_option
from messenger.constants import TransportMethod
from messenger.settings import ENVIRON_NAME, SETTINGS, Settings


class EchoTest(TestCase):
    def test_output(self):
        message = 'Smeghead!'
        output = 'Smeghead!\n'
        data = (
            # settings, verbosity, result
            (Verbosity.SILENT, Verbosity.SILENT, output),
            (Verbosity.SILENT, Verbosity.INFO, ''),
            (Verbosity.SILENT, Verbosity.DETAIL, ''),
            (Verbosity.SILENT, Verbosity.DEBUG, ''),
            (Verbosity.INFO, Verbosity.SILENT, output),
            (Verbosity.INFO, Verbosity.INFO, output),
            (Verbosity.INFO, Verbosity.DETAIL, ''),
            (Verbosity.INFO, Verbosity.DEBUG, ''),
            (Verbosity.DETAIL, Verbosity.SILENT, output),
            (Verbosity.DETAIL, Verbosity.INFO, output),
            (Verbosity.DETAIL, Verbosity.DETAIL, output),
            (Verbosity.DETAIL, Verbosity.DEBUG, ''),
            (Verbosity.DEBUG, Verbosity.SILENT, output),
            (Verbosity.DEBUG, Verbosity.INFO, output),
            (Verbosity.DEBUG, Verbosity.DETAIL, output),
            (Verbosity.DEBUG, Verbosity.DEBUG, output),
            (1, Verbosity.SILENT, output),
            (1, Verbosity.INFO, output),
            (1, Verbosity.DETAIL, ''),
            (1, Verbosity.DEBUG, ''),
        )

        for settings, verbosity, result in data:
            with self.subTest(settings=settings, verbosity=verbosity):
                SETTINGS['_verbosity'] = settings

                with OutputCapture(separate=True) as capture:
                    echo(message, verbosity)

                self.assertEqual(capture.stdout.getvalue(), result)
                self.assertEqual(capture.stderr.getvalue(), '')

    def test_error(self):
        SETTINGS['_verbosity'] = Verbosity.DEBUG

        with OutputCapture(separate=True) as capture:
            echo('Smeghead!', Verbosity.DEBUG, err=True)

        self.assertEqual(capture.stdout.getvalue(), '')
        self.assertEqual(capture.stderr.getvalue(), 'Smeghead!\n')

    def test_no_newline(self):
        SETTINGS['_verbosity'] = Verbosity.DEBUG

        with OutputCapture(separate=True) as capture:
            echo('Smeghead!', Verbosity.DEBUG, nl=False)

        self.assertEqual(capture.stdout.getvalue(), 'Smeghead!')
        self.assertEqual(capture.stderr.getvalue(), '')


@messenger_command()
def test_command():
    pass


class MessengerCommandTest(TestCase):
    def setUp(self):
        # mypy expects env to be Dict[str, str] for some reason
        self.runner = CliRunner(env={ENVIRON_NAME: None})  # type: ignore[dict-item]
        # Use new settings instance for tests.
        self.old_settings = SETTINGS.settings
        SETTINGS.settings = Settings()

    def tearDown(self):
        SETTINGS.settings = self.old_settings

    def test_help(self):
        result = self.runner.invoke(test_command, ['--help'])

        self.assertEqual(result.exit_code, 0)
        self.assertIn('--help', result.stdout)

    def test_version(self):
        result = self.runner.invoke(test_command, ['--version'])

        self.assertEqual(result.exit_code, 0)
        self.assertIn('fred-messenger ' + messenger.__version__, result.stdout)

    def test_config(self):
        db_connection = 'file://example/connection'
        config_file = tempfile.NamedTemporaryFile(mode='w')
        yaml.dump({'db_connection': db_connection}, config_file)

        result = self.runner.invoke(test_command, ['--config', config_file.name])

        self.assertEqual(SETTINGS['db_connection'], db_connection)
        self.assertEqual(result.exit_code, 0)

    def test_connection(self):
        db_connection = 'args://example/connection'

        result = self.runner.invoke(test_command, ['--connection', db_connection])

        self.assertEqual(SETTINGS['db_connection'], db_connection)
        self.assertEqual(result.exit_code, 0)

    def test_verbosity_long(self):
        result = self.runner.invoke(test_command, ['--verbosity', '3'])

        self.assertEqual(SETTINGS['_verbosity'], 3)
        self.assertEqual(result.exit_code, 0)

    def test_verbosity_short(self):
        result = self.runner.invoke(test_command, ['-v', '3'])

        self.assertEqual(SETTINGS['_verbosity'], 3)
        self.assertEqual(result.exit_code, 0)

    def test_group(self):
        # Test command can be set as a member of a click group.
        @click.group()
        def group():
            pass

        @messenger_command(command=group.command)
        def test_command():
            click.echo('Subgroup!')

        result = self.runner.invoke(group, ['test-command'])

        self.assertEqual(result.exit_code, 0)
        # In py36-postgres the output contains a spurious ResourceWarning.
        # Thus, we use assertIn instead of assertEqual to avoid that.
        self.assertIn('Subgroup!\n', result.stdout)


@click.command()
@type_option
def test_type_command(types: List[TransportMethod]) -> None:
    """Simple command to test type_option."""
    click.echo("TYPE: {}".format([t.value for t in types]))


class TypeOptionTest(TestCase):
    def setUp(self):
        self.runner = CliRunner()

    def test_default(self):
        result = self.runner.invoke(test_type_command)

        self.assertEqual(result.exit_code, 0)
        # In py38-sqlite and py39-sqlite the output contains a spurious ResourceWarning.
        # Thus, we use assertIn instead of assertEqual to avoid that.
        self.assertIn("TYPE: ['email', 'sms', 'letter']", result.stdout)

    def test_type_value(self):
        data = (
            ('all', ['email', 'sms', 'letter']),
            ('email', ['email']),
            ('sms', ['sms']),
            ('letter', ['letter']),
        )
        for type, output in data:
            with self.subTest(type=type):
                result = self.runner.invoke(test_type_command, ['--type', type])

                self.assertEqual(result.exit_code, 0)
                self.assertIn('TYPE: {}\n'.format(output), result.stdout)

    def test_invalid(self):
        result = self.runner.invoke(test_type_command, ['--type', 'invalid'])

        self.assertNotEqual(result.exit_code, 0)
        self.assertIn("Invalid value for '--type'", result.stdout)
